'''
Miscellaneous utility functions
'''

import sys
import os
import subprocess

def simplify(x):
    '''Take a list of tuples where each tuple is in form [v1,v2,...vn]
    and then coalesce all tuples tx and ty where tx[v1] equals ty[v2],
    preserving v3...vn of tx and discarding v3...vn of ty.

    m = [
    (1,2,"one"),
    (2,3,"two"),
    (3,4,"three"),
    (8,9,"three"),
    (4,5,"four"),
    (6,8,"blah"),
    ]
    simplify(x) -> [[1, 5, 'one'], [6, 9, 'blah']]
    '''
    y = list(x)
    if len(x) < 2: return y
    for idx,o in enumerate(list(y)):
        for idx2,p in enumerate(list(y)):
            if idx == idx2: continue
            if o and p and o[0] == p[1]:
                y[idx] = None
                y[idx2] = list(p)
                y[idx2][0] = p[0]
                y[idx2][1] = o[1]
    return [ n for n in y if n is not None ]

def uniq(seq, idfun=None):
    '''Makes a sequence 'unique' in the style of UNIX command uniq'''
    # order preserving
    if idfun is None:
        def idfun(x): return x
    seen = {}
    result = []
    for item in seq:
        marker = idfun(item)
        # in old Python versions:
        # if seen.has_key(marker)
        # but in new ones:
        if marker in seen: continue
        seen[marker] = 1
        result.append(item)
    return result


class SpecialPopen(subprocess.Popen):
    def __init__(self, *a, **kw):
        self._saved_args = a[0] if kw.get("args") is None else kw.get("args")
        subprocess.Popen.__init__(self, *a, **kw)


def progressbar(pipe, bufsize=-1, ratelimit=-1):

    def clpbar(cmdname):
        barargs = []
        if bufsize != -1:
            barargs = ["-bs", str(bufsize)]
        if ratelimit != -1:
            barargs = barargs + ['-th', str(ratelimit)]
        barprg = SpecialPopen(
            [cmdname, "-dan"] + barargs,
            stdin=pipe, stdout=subprocess.PIPE, bufsize=bufsize)
        return barprg

    def pv(cmdname):
        barargs = []
        if bufsize != -1:
            barargs = ["-B", str(bufsize)]
        if ratelimit != -1:
            barargs = barargs + ['-L', str(ratelimit)]
        barprg = SpecialPopen(
            [cmdname, "-ptrb"] + barargs,
            stdin=pipe, stdout=subprocess.PIPE, bufsize=bufsize)
        return barprg

    barprograms = [
        ("bar", clpbar),
        ("clpbar", clpbar),
        ("pv", pv),
    ]

    for name, func in barprograms:
        try:
            subprocess.call([name, '-h'], stdout=file(os.devnull, "w"), stderr=file(os.devnull, "w"), stdin=file(os.devnull, "r"))
        except OSError, e:
            if e.errno == 2: continue
            assert 0, "not reached while searching for clpbar or pv"
        return func(name)
    raise OSError(2, "no such file or directory searching for clpbar or pv")

def stderr(text):
    """print out something to standard error, followed by an ENTER"""
    sys.stderr.write(text)
    sys.stderr.write("\n")

__verbose = False
def verbose_stderr(*args, **kwargs):
    global __verbose
    if __verbose: stderr(*args, **kwargs)

def set_verbose(boolean):
    global __verbose
    __verbose = boolean



'''
Tree models for the ZFS tools
'''

from collections import OrderedDict

class Dataset(object):
    name = None
    children = None
    _properties = None
    parent = None
    invalidated = False
    def __init__(self, name, parent=None):
        self.name = name
        self.children = []
        self._properties = {}
        if parent:
            self.parent = parent
            self.parent.add_child(self)

    def add_child(self, child):
        self.children.append(child)
        return child

    def get_child(self, name):
        child = [ c for c in self.children if c.name == name and isinstance(c, Dataset) ]
        assert len(child) < 2
        if not child: raise KeyError, name
        return child[0]

    def get_snapshots(self, flt=True):
        if flt is True: flt = lambda x:True
        children = [ c for c in self.children if isinstance(c, Snapshot) and flt(c) ]
        return children

    def get_snapshot(self, name):
        children = [ c for c in self.get_snapshots() if c.name == name ]
        assert len(children) < 2
        if not children: raise KeyError, name
        return children[0]

    def lookup(self, name):  # FINISH THIS
        if "@" in name:
            path, snapshot = name.split("@")
        else:
            path = name
            snapshot = None

        if "/" not in path:
            try: dset = self.get_child(path)
            except KeyError: raise KeyError, "No such dataset %s at %s" % (path, self.get_path())
            if snapshot:
                try: dset = dset.get_snapshot(snapshot)
                except KeyError: raise KeyError, "No such snapshot %s at %s" % (snapshot, dset.get_path())
        else:
            head, tail = path.split("/", 1)
            try: child = self.get_child(head)
            except KeyError: raise KeyError, "No such dataset %s at %s" % (head, self.get_path())
            if snapshot: tail = tail + "@" + snapshot
            dset = child.lookup(tail)

        return dset

    def remove(self, child):
        if child not in self.children: raise KeyError, child.name
        child.invalidated = True
        child.parent = None
        self.children.remove(child)
        for c in child.children:
            child.remove(c)

    def get_path(self):
        if not self.parent: return self.name
        return "%s/%s" % (self.parent.get_path(), self.name)

    def get_relative_name(self):
        if not self.parent: return self.name
        return self.get_path()[len(self.parent.get_path()) + 1:]

    def walk(self):
        assert not self.invalidated, "%s invalidated" % self
        yield self
        for c in self.children:
            for element in c.walk():
                yield element

    def __iter__(self):
        return self.walk()

    def __str__(self):
        return "<Dataset:  %s>" % self.get_path()
    __repr__ = __str__

    def get_property(self,name):
        return self._properties[ name ]


class Pool(Dataset):
    def __str__(self):
        return "<Pool:     %s>" % self.get_path()
    __repr__ = __str__


class Snapshot(Dataset):
    # def __init__(self,name):
        # Dataset.__init__(self,name)
    def get_path(self):
        if not self.parent: return self.name
        return "%s@%s" % (self.parent.get_path(), self.name)

    def __str__(self):
        return "<Snapshot: %s>" % self.get_path()
    __repr__ = __str__


class PoolSet:  # maybe rewrite this as a dataset or something?
    pools = None

    def __init__(self):
        self.pools = {}

    def lookup(self, name):
        if "@" in name:
            path, snapshot = name.split("@")
        else:
            path = name
            snapshot = None

        if "/" not in path:
            try: dset = self.pools[path]
            except KeyError: raise KeyError, "No such pool %s" % (name)
            if snapshot:
                try: dset = dset.get_snapshot(snapshot)
                except KeyError: raise KeyError, "No such snapshot %s at %s" % (snapshot, dset.get_path())
        else:
            head, tail = path.split("/", 1)
            try: pool = self.pools[head]
            except KeyError: raise KeyError, "No such pool %s" % (head)
            if snapshot: tail = tail + "@" + snapshot
            dset = pool.lookup(tail)

        return dset

    def parse_zfs_r_output(self, creationtimes, properties = None ):
        """Parse the output of tab-separated zfs list.

        properties must be a list of property names expected to be found as
        tab-separated entries on each line of creationtimes after the
        dataset name and a tab.
        E.g. if properties passed here was ['creation'], we would expect
        each creationtimes line to look like 'dataset	3249872348'
        """
        properties = ['name', 'creation'] if properties == None else ['name'] + properties

        def extract_properties( line ):
            items = s.strip().split( '\t' )
            assert len( items ) == len( properties ), (properties, items)
            propvalues = map( lambda x: None if x == '-' else x, items[ 1: ] )
            return [ items[ 0 ], zip( properties[ 1: ], propvalues ) ]

        # make into array
        creations = OrderedDict([ extract_properties( s ) for s in creationtimes.splitlines() if s.strip() ])

        # names of pools
        old_dsets = [ x.get_path() for x in self.walk() ]
        old_dsets.reverse()
        new_dsets = creations.keys()

        for dset in new_dsets:
            if "@" in dset:
                dset, snapshot = dset.split("@")
            else:
                snapshot = None
            if "/" not in dset:  # pool name
                if dset not in self.pools:
                    self.pools[dset] = Pool(dset)
                    fs = self.pools[dset]
            poolname, pathcomponents = dset.split("/")[0], dset.split("/")[1:]
            fs = self.pools[poolname]
            for pcomp in pathcomponents:
                # traverse the child hierarchy or create if that fails
                try: fs = fs.get_child(pcomp)
                except KeyError:
                    fs = Dataset(pcomp, fs)

            if snapshot:
                if snapshot not in [ x.name for x in fs.children ]:
                    fs = Snapshot(snapshot, fs)

            fs._properties.update( creations[fs.get_path()] )

        for dset in old_dsets:
            if dset not in new_dsets:
                if "/" not in dset and "@" not in dset:  # a pool
                    self.remove(dset)
                else:
                    d = self.lookup(dset)
                    d.parent.remove(d)

    def remove(self, name):  # takes a NAME, unlike the child that is taken in the remove of the dataset method
        for c in self.pools[name].children:
            self.pools[name].remove(c)
        self.pools[name].invalidated = True
        del self.pools[name]

    def __getitem__(self, name):
        return self.pools[name]

    def __str__(self):
        return "<PoolSet at %s>" % id(self)
    __repr__ = __str__

    def walk(self):
        for item in self.pools.values():
            for dset in item.walk():
                yield dset

    def __iter__(self):
        return self.walk()



'''
Synchronization-related functionality
'''

import itertools
from zfstools.util import simplify
from zfstools.models import Snapshot
import warnings

# it is time to determine which datasets need to be synced
# we walk the entire dataset structure, and sync snapshots recursively
def recursive_replicate(s, d):
    sched = []

    # we first collect all snapshot names, to later see if they are on both sides, one side, or what
    all_snapshots = []
    if s: all_snapshots.extend(s.get_snapshots())
    if d: all_snapshots.extend(d.get_snapshots())
    all_snapshots = [ y[1] for y in sorted([ (x.get_property('creation'), x.name) for x in all_snapshots ]) ]
    snapshot_pairs = []
    for snap in all_snapshots:
        try: ssnap = s.get_snapshot(snap)
        except (KeyError, AttributeError): ssnap = None
        try: dsnap = d.get_snapshot(snap)
        except (KeyError, AttributeError): dsnap = None
        # if the source snapshot exists and is not already in the table of snapshots
        # then pair it up with its destination snapshot (if it exists) or None
        # and add it to the table of snapshots
        if ssnap and not snap in [ x[0].name for x in snapshot_pairs ]:
            snapshot_pairs.append((ssnap, dsnap))

    # now we have a list of all snapshots, paired up by name, and in chronological order
    # (it's quadratic complexity, but who cares)
    # now we need to find the snapshot pair that happens to be the the most recent common pair
    found_common_pair = False
    for idx, (m, n) in enumerate(snapshot_pairs):
        if m and n and m.name == n.name:
            found_common_pair = idx

    # we have combed through the snapshot pairs
    # time to check what the latest common pair is
    if not s.get_snapshots():
        if d is None:
            # well, no snapshots in source, just create a stub in the target
            sched.append(("create_stub", s, d, None, None))
    elif found_common_pair is False:
        # no snapshot is in common, problem!
        # theoretically destroying destination dataset and resyncing it recursively would work
        # but this requires work in the optimizer that comes later
        if d is not None and d.get_snapshots():
            warnings.warn("Asked to replicate %s into %s but %s has snapshots and both have no snapshots in common!" % (s, d, d))
        # see source snapshots
        full_source_snapshots = [ y[1] for y in sorted([ (x.get_property('creation'), x) for x in s.get_snapshots() ]) ]
        # send first snapshot as full snapshot
        sched.append(("full", s, d, None, full_source_snapshots[0]))
        if len(full_source_snapshots) > 1:
            # send other snapshots as incremental snapshots
            sched.append(("incremental", s, d, full_source_snapshots[0], full_source_snapshots[-1]))
    elif found_common_pair == len(snapshot_pairs) - 1:
        # the latest snapshot of both datasets that is common to both, is the latest snapshot in the source
        # we have nothing to do here because the datasets are "in sync"
        pass
    else:
        # the source dataset has more recent snapshots, not present in the destination dataset
        # we need to transfer those
        snapshots_to_transfer = [ x[0] for x in snapshot_pairs[found_common_pair:] ]
        for n, x in enumerate(snapshots_to_transfer):
            if n == 0: continue
            sched.append(("incremental", s, d, snapshots_to_transfer[n - 1], x))

    # now let's apply the same argument to the children
    children_sched = []
    for c in [ x for x in s.children if not isinstance(x, Snapshot) ]:
        try: cd = d.get_child(c.name)
        except (KeyError, AttributeError): cd = None
        children_sched.extend(recursive_replicate(c, cd))

    # and return our schedule of operations to the parent
    return sched + children_sched

def optimize_coalesce(operation_schedule):
    # now let's optimize the operation schedule
    # this optimization is quite basic
    # step 1: coalesce contiguous operations on the same file system

    operations_grouped_by_source = itertools.groupby(
        operation_schedule,
        lambda op: op[1]
    )
    new = []
    for _, opgroup in [ (x, list(y)) for x, y in operations_grouped_by_source ]:
        if not opgroup:  # empty opgroup
            continue
        if opgroup[0][0] == 'full':  # full operations
            new.extend(opgroup)
        elif opgroup[0][0] == 'create_stub':  # create stub operations
            new.extend(opgroup)
        elif opgroup[0][0] == 'incremental':  # incremental
            # 1->2->3->4 => 1->4
            new_ops = [ (srcs, dsts) for _, _, _, srcs, dsts in opgroup ]
            new_ops = simplify(new_ops)
            for srcs, dsts in new_ops:
                new.append(tuple(opgroup[0][:3] + (srcs, dsts)))
        else:
            assert 0, "not reached: unknown operation type in %s" % opgroup
    return new

def optimize_recursivize(operation_schedule):
    def recurse(dataset, func):
        results = []
        results.append((dataset, func(dataset)))
        results.extend([ x for child in dataset.children if child.__class__ != Snapshot for x in recurse(child, func) ])
        return results

    def zero_out_sched(dataset):
        dataset._ops_schedule = []

    def evict_sched(dataset):
        dataset._ops_schedule = []

    operations_grouped_by_source = itertools.groupby(
        operation_schedule,
        lambda op: op[1]
    )
    operations_grouped_by_source = [ (x, list(y)) for x, y in operations_grouped_by_source ]

    roots = set()
    for root, opgroup in operations_grouped_by_source:
        while root.parent is not None:
            root = root.parent
        roots.add(root)

    for root in roots:
        recurse(root, zero_out_sched)

    for source, opgroup in operations_grouped_by_source:
        source._ops_schedule = opgroup

    def compare(*ops_schedules):
        assert len(ops_schedules), "operations schedules cannot be empty: %r" % ops_schedules

        # in the case of the list of operations schedules being just one (no children)
        # we return True, cos it's safe to recursively replicate this one
        if len(ops_schedules) == 1:
            return True

        # now let's check that all ops schedules are the same length
        # otherwise they are not the same and we can say the comparison isn't the same
        lens = set([ len(o) for o in ops_schedules ])
        if len(lens) != 1:
            return False

        # we have multiple schedules
        # if their type, snapshot origin and snapshot destination are all the same
        # we can say that they are "the same"
        comparisons = [
                all([
                    # never attempt to recursivize operations who involve create_stub
                    all(["create_stub" not in o[0] for o in ops]),
                    len(set([o[0] for o in ops])) == 1,
                    any([o[3] is None for o in ops]) or len(set([o[3].name for o in ops])) == 1,
                    any([o[4] is None for o in ops]) or len(set([o[4].name for o in ops])) == 1,
                ])
                for ops
                in zip(*ops_schedules)
        ]
        return all(comparisons)

    # remove unnecessary stubs that stand in for only other stubs
    for root in roots:
        for dataset, _ in recurse(root, lambda d: d):
            ops = [z for x, y in recurse(dataset, lambda d: d._ops_schedule) for z in y]
            if all([o[0] == 'create_stub' for o in ops]):
                dataset._ops_schedule = []

    for root in roots:
        for dataset, _ in recurse(root, lambda d: d):
            if compare(*[y for x, y in recurse(dataset, lambda d: d._ops_schedule)]):
                old_ops_schedule = dataset._ops_schedule
                recurse(dataset, zero_out_sched)
                for op in old_ops_schedule:
                    dataset._ops_schedule.append((
                        op[0] + "_recursive", op[1], op[2], op[3], op[4]
                    ))

    new_operation_schedule = []
    for root in roots:
        for dataset, ops_schedule in recurse(root, lambda d: d._ops_schedule):
            new_operation_schedule.extend(ops_schedule)

    for root in roots:
        recurse(root, evict_sched)

    return new_operation_schedule

def optimize(operation_schedule, allow_recursivize = True):
    operation_schedule = optimize_coalesce(operation_schedule)
    if allow_recursivize:
        operation_schedule = optimize_recursivize(operation_schedule)
    return operation_schedule

# we walk the entire dataset structure, and sync snapshots recursively
def recursive_clear_obsolete(s, d):
    sched = []

    # we first collect all snapshot names, to later see if they are on both sides, one side, or what
    all_snapshots = []
    snapshots_in_src = set([ m.name for m in s.get_snapshots() ])
    snapshots_in_dst = set([ m.name for m in d.get_snapshots() ])

    snapshots_to_delete = snapshots_in_dst - snapshots_in_src
    snapshots_to_delete = [ d.get_snapshot(m) for m in snapshots_to_delete ]

    for m in snapshots_to_delete:
        sched.append(("destroy", m))

    # now let's apply the same argument to the children
    children_sched = []
    for child_d in [ x for x in d.children if not isinstance(x, Snapshot) ]:
        child_s = None

        try:
            child_s = s.get_child(child_d.name)
        except (KeyError, AttributeError):
            children_sched.append(("destroy_recursively", child_d))

        if child_s:
            children_sched.extend(recursive_clear_obsolete(child_s, child_d))

    # and return our schedule of operations to the parent
    return sched + children_sched



'''
ZFS connection classes
'''

import subprocess
import os
from zfstools.models import PoolSet
from zfstools.util import progressbar, SpecialPopen, verbose_stderr
from Queue import Queue
from threading import Thread


# Work-around for check_output not existing on Python 2.6, as per
# http://stackoverflow.com/questions/4814970/subprocess-check-output-doesnt-seem-to-exist-python-2-6-5
# The implementation is lifted from
# http://hg.python.org/cpython/file/d37f963394aa/Lib/subprocess.py#l544
if "check_output" not in dir( subprocess ): # duck punch it in!
    def f(*popenargs, **kwargs):
        if 'stdout' in kwargs:
            raise ValueError('stdout argument not allowed, it will be overridden.')
        process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd) # , output=output)
        return output
    subprocess.check_output = f

class ZFSConnection:
    host = None
    _poolset = None
    _dirty = True
    _trust = False
    _properties = None
    def __init__(self,host="localhost", trust=False, sshcipher=None, properties=None, identityfile=None, knownhostsfile=None, verbose=False):
        self.host = host
        self._trust = trust
        self._properties = properties if properties else []
        self._poolset= PoolSet()
        self.verbose = verbose
        if host in ['localhost','127.0.0.1']:
            self.command = []
        else:
            self.command = ["ssh","-o","BatchMode=yes","-a","-x"]
            if self._trust:
                self.command.extend(["-o","CheckHostIP=no"])
                self.command.extend(["-o","StrictHostKeyChecking=no"])
            if sshcipher != None:
                self.command.extend(["-c",sshcipher])
            if identityfile != None:
                self.command.extend(["-i",identityfile])
            if knownhostsfile != None:
                self.command.extend(["-o","UserKnownHostsFile=%s" % knownhostsfile])
            self.command.extend([self.host])

    def _get_poolset(self):
        if self._dirty:
            properties = [ 'creation' ] + self._properties
            stdout2 = subprocess.check_output(self.command + ["zfs", "list", "-Hpr", "-o", ",".join( ['name'] + properties ), "-t", "all"])
            self._poolset.parse_zfs_r_output(stdout2,properties)
            self._dirty = False
        return self._poolset
    pools = property(_get_poolset)

    def create_dataset(self, name, parents=False):
        parents_opt = ["-p"] if parents else []
        subprocess.check_call(self.command + ["zfs", "create"] + parents_opt + [name])
        self._dirty = True
        return self.pools.lookup(name)

    def destroy_dataset(self, name):
        subprocess.check_call(self.command + ["zfs", "destroy", name])
        self._dirty = True

    def destroy_recursively(self, name, returnok=False):
        """If returnok, then simply return success as a boolean."""
        ok = True
        cmd = self.command + ["zfs", "destroy", '-r', name]
        if returnok:
            ok = subprocess.call(cmd) == 0
        else:
            subprocess.check_call(cmd)
        self._dirty = True
        return ok

    def snapshot_recursively(self,name,snapshotname,properties={}):
        plist = sum( map( lambda x: ['-o', '%s=%s' % x ], properties.items() ), [] )
        subprocess.check_call(self.command + ["zfs", "snapshot", "-r" ] + plist + [ "%s@%s" % (name, snapshotname)])
        self._dirty = True

    def send(self,name,opts=None,bufsize=-1,compression=False,lockdataset=None):
        if not opts: opts = []
        cmd = list(self.command)
        if compression and cmd[0] == 'ssh': cmd.insert(1,"-C")
        if lockdataset is not None:
            cmd += ["zflock"]
            if self.verbose:
                cmd += ["-v"]
            cmd += [lockdataset, "--"]
        cmd += ["zfs", "send"] + opts
        if "-t" not in opts:
            # if we're resuming, don't specify the name of what to send
            cmd += [name]
        verbose_stderr("%s\n" % ' '.join(cmd))
        p = SpecialPopen(cmd,stdin=file(os.devnull),stdout=subprocess.PIPE,bufsize=bufsize)
        return p

    def receive(self,name,pipe,opts=None,bufsize=-1,compression=False,lockdataset=None):
        if not opts: opts = []
        cmd = list(self.command)
        if compression and cmd[0] == 'ssh': cmd.insert(1,"-C")
        if lockdataset is not None:
            cmd += ["zflock"]
            if self.verbose:
                cmd += ["-v"]
            cmd += [lockdataset, "--"]
        cmd = cmd + ["zfs", "receive"] + opts + [name]
        verbose_stderr("%s\n" % ' '.join(cmd))
        p = SpecialPopen(cmd,stdin=pipe,bufsize=bufsize)
        return p

    def transfer(self, dst_conn, s, d, fromsnapshot=None, showprogress=False, bufsize=-1, send_opts=None, receive_opts=None, ratelimit=-1, compression=False, locksrcdataset=None, lockdstdataset=None, verbose=False, resumable=False):
        if send_opts is None: send_opts = []
        if receive_opts is None: receive_opts = []

        try:
            resume_token = dst_conn.pools.lookup(d).get_property("receive_resume_token")
        except:
            resume_token = None

        queue_of_killables = Queue()

        if fromsnapshot: fromsnapshot=["-i",fromsnapshot]
        else: fromsnapshot = []
        if verbose: verbose_opts = ["-v"]
        else: verbose_opts = []
        # Regardless of whether resumable is requested this time , if
        # there's a resume token on the destination, we have to use it.
        if resume_token is not None:
            all_send_opts = ["-t", resume_token] + verbose_opts
        else:
            all_send_opts = [] + fromsnapshot + send_opts + verbose_opts
        sndprg = self.send(s, opts=all_send_opts, bufsize=bufsize, compression=compression, lockdataset=locksrcdataset)
        sndprg_supervisor = Thread(target=lambda: queue_of_killables.put((sndprg, sndprg.wait())))
        sndprg_supervisor.start()

        if showprogress:
            try:
                        barprg = progressbar(pipe=sndprg.stdout,bufsize=bufsize,ratelimit=ratelimit)
                        barprg_supervisor = Thread(target=lambda: queue_of_killables.put((barprg, barprg.wait())))
                        barprg_supervisor.start()
                        sndprg.stdout.close()
            except OSError:
                        os.kill(sndprg.pid,15)
                        raise
        else:
            barprg = sndprg

        try:
                        if resumable: resumable_recv_opts = ["-s"]
                        else: resumable_recv_opts = []
                        all_recv_opts = ["-Fu"] + verbose_opts + resumable_recv_opts + receive_opts
                        rcvprg = dst_conn.receive(d,pipe=barprg.stdout,opts=all_recv_opts,bufsize=bufsize,compression=compression, lockdataset=lockdstdataset)
                        rcvprg_supervisor = Thread(target=lambda: queue_of_killables.put((rcvprg, rcvprg.wait())))
                        rcvprg_supervisor.start()
                        barprg.stdout.close()
        except OSError:
                os.kill(sndprg.pid, 15)
                if sndprg.pid != barprg.pid: os.kill(barprg.pid, 15)
                raise

        dst_conn._dirty = True
        allprocesses = set([rcvprg, sndprg]) | ( set([barprg]) if showprogress else set() )
        while allprocesses:
            diedprocess, retcode = queue_of_killables.get()
            allprocesses = allprocesses - set([diedprocess])
            if retcode != 0:
                [ p.kill() for p in allprocesses ]
                raise subprocess.CalledProcessError(retcode, diedprocess._saved_args)



'''
Tests for the zfstools.util module
'''
import unittest
from zfstools import util as zfsutils


class TestSimplify(unittest.TestCase):

    def test_simple(self):
        m = [
        (1,2,"one"),
        (2,3,"two"),
        ]
        r1 = zfsutils.simplify(m)
        r2 = [[1, 3, 'one']]
        self.assertEquals(r1, r2)        

    def test_complex(self):
        m = [
        (1,2,"one"),
        (2,3,"two"),
        (3,4,"three"),
        (8,9,"three"),
        (4,5,"four"),
        (6,8,"blah"),
        ]
        r1 = zfsutils.simplify(m)
        r2 = [[1, 5, 'one'], [6, 9, 'blah']]
        self.assertEquals(r1, r2)        

    def test_discrete(self):
        m = [
        (1,2,"one"),
        (2,4,"two"),
        (6,9,"three"),
        ]
        # note last element is a tuple
        r1 = zfsutils.simplify(m)
        r2 = [[1, 4, 'one'], (6, 9, 'three')]
        self.assertEquals(r1, r2)        

    def test_with_strings(self):
        m = [
        "abM",
        "bcN",
        "cdO",
        ]
        # note last element is a tuple
        r1 = zfsutils.simplify(m)
        r2 = [list("adM")]
        self.assertEquals(r1, r2)        

class TestUniq(unittest.TestCase):
    
    def test_identity(self):
        s = "abc"
        r1 = zfsutils.uniq(s)
        r2 = list(s)
        self.assertEquals(r1,r2)

    def test_similarelement(self):
        s = "abcb"
        r1 = zfsutils.uniq(s)
        r2 = list("abc")
        self.assertEquals(r1,r2)

    def test_repeatedsequences(self):
        s = "abcabc"
        r1 = zfsutils.uniq(s)
        r2 = list("abc")
        self.assertEquals(r1,r2)

    def test_idfun(self):
        s = "abc"
        idfun = lambda _: "a"
        r1 = zfsutils.uniq(s, idfun)
        r2 = list("a")
        self.assertEquals(r1,r2)

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
